#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "srv.h"

bool test = 0;

const char *old_s, *new_s;
int old, new;

char *usage =
"Usage:\n"
"  runlevelchange [--test] [prevlevel runlevel]\n"
"prevlevel is one of N12345\n"
"runlevel is one of S0123456\n"
"If prevlevel and runlevel are not given, they are taken from their\n"
"uppercase equivalent environment variables.";

static int whattodo (const struct service *service)
{
	int ret = 0;
	if (old >= 1 && old <= 5 && service->runlevel[old]) ret--;
	if (new >= 1 && new <= 5 && service->runlevel[new]) ret++;
	return ret;
}

static int sortorder (const void *v1, const void *v2)
{
	const struct service *s1 = v1;
	const struct service *s2 = v2;
	int i1, i2;
	i1 = whattodo(s1);
	i2 = whattodo(s2);
	if (i1 != i2) return i1 - i2;
	if (s1->sequence != s2->sequence) {
		if (i1 < 0) return s2->sequence - s1->sequence;
		else return s1->sequence - s2->sequence;
	}
	if (i1 < 0) return strcmp(s2->name, s1->name);
	else return strcmp(s1->name, s2->name);
}

int main (int argc, char *argv[])
{
	int i;
	struct service *service;
	struct services services;

	if (argc >= 2 && !strcmp(argv[1], "--test")) {
		test = 1;
		argv++;
		argc--;
	}

	if (argc == 1) {
		old_s = getenv("PREVLEVEL");
		new_s = getenv("RUNLEVEL");
	} else if (argc == 3) {
		old_s = argv[1];
		new_s = argv[2];
	} else {
		message(usage);
		exit(1);
	}
	if (old_s == NULL || new_s == NULL ||
	  strlen(old_s) != 1 || strlen(new_s) != 1 ||
	  strchr("N12345", old_s[0]) == NULL ||
	  strchr("S0123456", new_s[0]) == NULL) {
		message(usage);
		exit(1);
	}

	if (old_s[0] == 'N') old = -2;
	else old = old_s[0] - '0';
	if (new_s[0] == 'S') new = -1;
	else new = new_s[0] - '0';

	if (old == -2 && new == -1) {
		run_parts(BOOTDIR, NULL, test);
		return 0;
	}

	if (services_read_list(&services) < 0) exit(1);
	service = malloc(services.nr * sizeof(struct service));
	for (i = 0; i < services.nr; i++) {
		service_read(services.name[i], &service[i]);
	}

	qsort(service, services.nr, sizeof(struct service), sortorder);

	for (i = 0; i < services.nr; i++) {
		switch (whattodo(&service[i])) {
		  case -1:
			service_stop(&service[i], test);
			break;
		  case 1:
			service_start(&service[i], test);
			break;
		}
	}

	if (new == 0 || new == 6) {
		run_parts(HALTDIR, new ? "reboot" : "halt", test);
	}

	return 0;
}
